TK_SYS_ACK = 0xF0100001
TK_SYS_NAK = 0xF0100002
TK_SYS_BAD_REQUEST = 0xF0100003
TK_SYS_APPLICATION_ERROR = 0xF0100004
TK_SYS_QUEUE_FULL = 0xF0100005
TK_SYS_HEALTH_CHECK = 0xF0100006

TK_HV_FIND_SERVICE = 0x00200005
TK_HV_START_COMP = 0x00200007
TK_HV_SERVICEONLINE = 0x00200008
TK_HV_SERVICEOFFLINE = 0x00200009

TK_CMP_TERM_NOW = 0xF0300001
TK_CMP_GETSERVICES = 0xF0300002
TK_CMP_GETDEPENDS = 0xF0300003
TK_CMP_REGSERVICE = 0xF0300004
TK_CMP_START = 0xF0300005
TK_CMP_RELOAD = 0xF0300007
TK_CMP_RECREATE = 0xF0300008
TK_CMP_HEALTH_CHECK = 0xF0300009

TK_EVT_EVENT = 0xF0400001
TK_EVT_SUBSCRIBE = 0xF0400002
TK_EVT_UNSUBSCRIBE = 0xF0400003



from typing import Optional

from ._format import FM_UNDEFINED

TCP_MSG_TRAILER = 0x81010077


class MBMessage(object):
    def __init__(
        self,
        token: int,
        format: int = FM_UNDEFINED,
        ttl: int = -1,
        data: bytes = b"",
        headers: bytes = b"",
        trailer: int = TCP_MSG_TRAILER,
        destname: Optional[str] = None,
        remotehost: Optional[str] = None,
    ):
        self.token = token
        self.format = format
        self.ttl = ttl
        self.data = data
        self.headers = headers
        self.trailer = trailer
        self.destname = destname
        self.remotehost = remotehost



from threading import Thread

import cmsgbus
from assertpy import (
    assert_that,
)
from messagebus import (
    Event,
    Message,
)
from msgbus import (
    MBContext,
    MBService,
)


def test_event_published_from_python_is_received_on_msgbus(
    hv_address,
    hv_port,
):
    publisher = MBContext(
        my_service=MBService(
            address="127.0.0.1",
            port=14580,
        ),
        name="Publisher",
    )

    c_subscriber = cmsgbus.MBContext(
        my_service=cmsgbus.MBService(
            address="127.0.0.1",
            port=14581,
        ),
        name="Subscriber",
    )

    c_hv_service = cmsgbus.MBService(
        address=hv_address,
        port=hv_port,
    )

    hv_service = MBService(
        address=hv_address,
        port=hv_port,
    )

    subject = "MY_EVENT"
    event_type = "MY_TYPE"
    event_data = "EVENT_DATA"

    def publisher_thread():

        publisher.MB_EvtSend(
            hv_service=hv_service,
            subject=subject,
            type=event_type,
            xml=event_data,
            synchronous=False,
        )

    c_subscriber.MB_EvtSubscribe(
        hv_service=c_hv_service,
        subject=subject,
    )

    t = Thread(target=publisher_thread)
    t.start()

    event = c_subscriber.MB_GetMessage()
    event = Event.build_from_message(
        message=Message(
            token=event.token,
            data_type=event.format,
            data=event.data,
        ),
    )
    assert_that(event.subject).is_equal_to(subject)
    assert_that(event.event_type).is_equal_to(event_type)
    assert_that(event.data).is_equal_to(event_data.encode())
    assert_that(event.synchronous).is_false()

    t.join()

    publisher.MB_Finalize()
    c_subscriber.MB_Finalize()


def test_event_published_from_msgbus_is_received_on_python(
    hv_address,
    hv_port,
):
    c_publisher = cmsgbus.MBContext(
        my_service=cmsgbus.MBService(
            address="127.0.0.1",
            port=14580,
        ),
        name="Publisher",
    )

    subscriber = MBContext(
        my_service=MBService(
            address="127.0.0.1",
            port=14581,
        ),
        name="Subscriber",
    )

    c_hv_service = cmsgbus.MBService(
        address=hv_address,
        port=hv_port,
    )

    hv_service = MBService(
        address=hv_address,
        port=hv_port,
    )

    subject = "MY_EVENT"
    event_type = "MY_TYPE"
    event_data = "EVENT_DATA"

    def publisher_thread():

        c_publisher.MB_EvtSend(
            hv_service=c_hv_service,
            subject=subject,
            type=event_type,
            xml=event_data,
            synchronous=False,
        )

    subscriber.MB_EvtSubscribe(
        hv_service=hv_service,
        subject=subject,
    )

    t = Thread(target=publisher_thread)
    t.start()

    event = subscriber.MB_GetMessage()
    event = Event.build_from_message(
        message=Message(
            token=event.token,
            data_type=event.format,
            data=event.data,
        ),
    )
    assert_that(event.subject).is_equal_to(subject)
    assert_that(event.event_type).is_equal_to(event_type)
    assert_that(event.data).is_equal_to(event_data.encode())
    assert_that(event.synchronous).is_false()

    t.join()

    c_publisher.MB_Finalize()
    subscriber.MB_Finalize()

def test_event_published_from_python_is_received_on_python(
    hv_address,
    hv_port,
):
    publisher = MBContext(
        my_service=MBService(
            address="127.0.0.1",
            port=14580,
        ),
        name="Publisher",
    )

    subscriber = MBContext(
        my_service=MBService(
            address="127.0.0.1",
            port=14581,
        ),
        name="Subscriber",
    )

    hv_service = MBService(
        address=hv_address,
        port=hv_port,
    )

    subject = "MY_EVENT"
    event_type = "MY_TYPE"
    event_data = "EVENT_DATA"

    def publisher_thread():

        publisher.MB_EvtSend(
            hv_service=hv_service,
            subject=subject,
            type=event_type,
            xml=event_data,
            synchronous=False,
        )

    subscriber.MB_EvtSubscribe(
        hv_service=hv_service,
        subject=subject,
    )

    t = Thread(target=publisher_thread)
    t.start()

    event = subscriber.MB_GetMessage()
    event = Event.build_from_message(
        message=Message(
            token=event.token,
            data_type=event.format,
            data=event.data,
        ),
    )
    assert_that(event.subject).is_equal_to(subject)
    assert_that(event.event_type).is_equal_to(event_type)
    assert_that(event.data).is_equal_to(event_data.encode())
    assert_that(event.synchronous).is_false()

    t.join()

    publisher.MB_Finalize()
    subscriber.MB_Finalize()



import logging
import queue
import socket
import struct
import sys
from threading import (
    Lock,
    Thread,
)
from time import (
    sleep,
    time,
)
from typing import (
    Optional,
)

from ._MBException import MBException
from ._MBMessage import MBMessage
from ._MBService import MBService
from ._MBTimeout import MBTimeout
from ._crc import calc_crc
from ._format import (
    FM_PARAM,
    FM_STRING,
    FM_STR_ARRAY_PIPE,
    FM_UNDEFINED,
)
from ._status import (
    SE_MB_ERR,
    SE_MB_RCVBODY,
    SE_MB_RCVHDR,
    SE_MB_RCVTRAILER,
    SE_NOTFOUND,
    SE_SUCCESS,
    SE_TIMEOUT,
)
from ._token import (
    TK_CMP_GETDEPENDS,
    TK_CMP_GETSERVICES,
    TK_CMP_REGSERVICE,
    TK_CMP_START,
    TK_CMP_TERM_NOW,
    TK_EVT_EVENT,
    TK_EVT_SUBSCRIBE,
    TK_HV_FIND_SERVICE,
    TK_HV_START_COMP,
    TK_SYS_ACK,
    TK_SYS_NAK,
    TK_SYS_QUEUE_FULL,
    TK_EVT_UNSUBSCRIBE,
)

TCP_MSG_MAGIC_EXT = 0x21A2132F
TCP_MSG_TRAILER = 0x81010077
TCP_MSG_TRAILER_OW = 0x18989933

logger = logging.getLogger(__name__)


class MBContext(object):
    def __init__(
        self,
        my_service: Optional[MBService],
        name: str,
    ):
        self.my_service = my_service
        self.name = name

        if self.my_service is not None:
            self.msg_queue: queue.Queue[MBMessage] = queue.Queue(500)
            self.global_queue_lock = Lock()

            self.receive_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            linger_data = struct.pack('ii', 0, 0)
            self.receive_sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, linger_data)
            self.receive_sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self.receive_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.receive_sock.bind((my_service.address, my_service.port))
            self.receive_sock.listen(100)

            self.running = True
            self.receive_thread = Thread(
                target=self._receiver,
                name="MBContextReceiverThread",
                daemon=True,
            )
            self.receive_thread.start()

    def MB_GetMessage(self, filtermin=0, filtermax=0, timeout:Optional[int]=None) -> MBMessage:
        if filtermin == 0 and filtermax == 0:
            msg = self.msg_queue.get(timeout=timeout)
            logger.debug("Returning msg with token [0x%08X]", msg.token)
            return msg

        while True:
            self.global_queue_lock.acquire(blocking=True)
            all_messages = []
            try:
                while True:
                    all_messages.append(self.msg_queue.get(block=False))
            except queue.Empty:
                pass

            if len(all_messages) == 0:
                self.global_queue_lock.release()
                all_messages.append(self.msg_queue.get(block=True))
                self.global_queue_lock.acquire(blocking=True)
                try:
                    all_messages.append(self.msg_queue.get(block=False))
                except queue.Empty:
                    pass

            ret = None
            for msg in all_messages:
                if ret is None and filtermin <= msg.token <= filtermax:
                    ret = msg
                else:
                    self.msg_queue.put(msg)

            self.global_queue_lock.release()

            if ret is not None:
                logger.debug("Returning msg with token [0x%08X]", ret.token)
                return ret

    def MB_SendFullMessage(
        self,
        destination: MBService,
        token: int,
        format: int = 0,
        data: bytes = None,
        headers: bytes = None,
        timeout: int = -1,
        ttl: int = -1,
    ):

        return self._send_message(
            destination=destination,
            token=token,
            format=format,
            data=data,
            headers=headers,
            timeout=timeout,
            ttl=ttl,
            one_way=False,
        )

    def MB_SendMessage(
        self,
        destination: MBService,
        token: int,
        format: int = 0,
        data: Optional[bytes] = None,
        timeout: int = -1,
    ) -> MBMessage:

        return self.MB_SendFullMessage(
            destination=destination,
            token=token,
            format=format,
            data=data,
            timeout=timeout,
        )

    def MB_SendFullOneWayMessage(
        self,
        destination: MBService,
        token: int,
        format: int = 0,
        data: bytes = None,
        headers: bytes = None,
        timeout: int = -1,
        ttl: int = -1,
    ):
        self._send_message(
            destination=destination,
            token=token,
            format=format,
            data=data,
            headers=headers,
            timeout=timeout,
            ttl=ttl,
            one_way=True,
        )

    def MB_SendOneWayMessage(
        self,
        destination: MBService,
        token: int,
        format: int = 0,
        data: bytes = None,
        timeout: int = -1,
    ):
        self._send_message(
            destination=destination,
            token=token,
            format=format,
            data=data,
            headers=None,
            timeout=timeout,
            ttl=-1,
            one_way=True,
        )

    def MB_ReplyMessage(
        self,
        message: MBMessage,
        format: int = FM_UNDEFINED,
        data: bytes = b"",
    ) -> None:

        if message.trailer == TCP_MSG_TRAILER_OW:
            logger.error("MBMessage is one-way, must not reply")
            return

        sock = getattr(message, "_socket", None)
        if sock is None:
            logger.error("MBMessage without socket, already replied")
            return

        try:
            self._send_data(
                sock=sock,
                dest_service_name=None,
                token=message.token,
                format=format,
                data=data,
                headers=b"",
                ttl=-1,
                one_way=False,
                end_time=int(time() + 5),
            )
            sock.close()
        except (Exception,):
            logger.exception("Error sending reply")

        delattr(message, "_socket")

    def MB_EvtSubscribe(
        self,
        hv_service: MBService,
        subject: str,
    ) -> None:

        reply = self.MB_SendMessage(
            destination=hv_service,
            token=TK_EVT_SUBSCRIBE,
            format=FM_PARAM,
            data=f"{subject}\0{self.my_service.address}\0{self.my_service.port}".encode(),
        )

        if reply.token != TK_SYS_ACK:
            raise MBException("Error calling MB_EvtSubscribe.", SE_MB_ERR)

    def MB_EvtUnsubscribe(
        self,
        hv_service: MBService,
        subject: str,
    ) -> None:

        reply = self.MB_SendMessage(
            destination=hv_service,
            token=TK_EVT_UNSUBSCRIBE,
            format=FM_PARAM,
            data=f"{subject}\0{self.my_service.address}\0{self.my_service.port}".encode(),
        )

        if reply.token != TK_SYS_ACK:
            raise MBException("Error calling MB_EvtSubscribe.", SE_MB_ERR)

    def MB_EvtSend(
        self,
        hv_service: MBService,
        subject: str,
        type: str,
        xml: str,
        synchronous: bool = False,
        sourceid: int = 0,
        queue: str = None,
        timeout: int = -1,
    ):
        return self.MB_EvtSendFull(
            hv_service=hv_service,
            headers=None,
            subject=subject,
            type=type,
            xml=xml,
            synchronous=synchronous,
            sourceid=sourceid,
            queue=queue,
            timeout=timeout,
            ttl=-1,
        )

    def MB_EvtSendFull(
        self,
        hv_service: MBService,
        headers: Optional[bytes],
        subject: str,
        type: str,
        xml: str,
        synchronous: bool = False,
        sourceid: int = 0,
        queue: str = None,
        timeout: int = -1,
        ttl: int = -1,
    ):
        sync = "true" if synchronous else "false"
        timeout_in_seconds = timeout / 1000000
        evt_data = f"{xml}\0{subject}\0{type}\0{sync}\0{sourceid}\0{timeout_in_seconds}"
        if queue is not None:
            evt_data += f"\0{queue}"
        if not synchronous:
            self.MB_SendFullOneWayMessage(
                destination=hv_service,
                token=TK_EVT_EVENT,
                format=FM_PARAM,
                data=evt_data.encode(),
                headers=headers,
                timeout=timeout,
                ttl=ttl,
            )
            return MBMessage(token=TK_SYS_ACK)
        else:
            return self.MB_SendMessage(
                destination=hv_service,
                token=TK_EVT_EVENT,
                format=FM_PARAM,
                data=evt_data.encode(),
                timeout=timeout,
            )

    def MB_LocateService(
        self,
        hv_service: MBService,
        service_name: str,
        maxretries: int = 5,
    ):
        reply = self.MB_SendFullMessage(
            destination=hv_service,
            token=TK_HV_FIND_SERVICE,
            format=FM_STRING,
            data=service_name.encode(),
        )

        if reply.token == TK_HV_START_COMP and maxretries > 0:
            sleep(0.5)
            reply = self.MB_LocateService(
                hv_service=hv_service,
                service_name=service_name,
                maxretries=maxretries - 1,
            )

        if reply.token != TK_SYS_ACK:
            raise MBException("Error calling MB_LocateService.", SE_NOTFOUND)

        parts = reply.data.decode().split("\0")
        return MBService(
            address=parts[0],
            port=int(parts[1]),
        )

    def MB_RegisterService(
        self,
        my_service: MBService,
        hv_service: MBService,
        service_type: str,
        service_name: str,
    ) -> None:

        reply = self.MB_SendFullMessage(
            destination=hv_service,
            token=TK_CMP_REGSERVICE,
            format=FM_PARAM,
            data=f"{service_type}\0{service_name}\0{my_service.address}\0{my_service.port}".encode(),
        )

        if reply.token != TK_SYS_ACK:
            raise MBException("Error calling MB_RegisterService.", SE_MB_ERR)

    def MB_WaitStart(
        self,
        exported_services: Optional[str] = None,
        required_services: Optional[str] = None,
    ) -> int:
        while True:
            msg = self.MB_GetMessage(filtermin=TK_CMP_TERM_NOW, filtermax=TK_CMP_START)
            if msg.token == TK_CMP_TERM_NOW:
                msg.token = TK_SYS_ACK
                self.MB_ReplyMessage(
                    message=msg,
                    format=FM_UNDEFINED,
                    data=b"",
                )
                raise MBException("Component is terminating", SE_MB_ERR)

            elif msg.token == TK_CMP_GETSERVICES:
                if exported_services is not None and exported_services != "":
                    msg.token = TK_SYS_ACK
                    self.MB_ReplyMessage(
                        message=msg,
                        format=FM_STR_ARRAY_PIPE,
                        data=exported_services.encode(),
                    )
                else:
                    msg.token = TK_SYS_NAK
                    self.MB_ReplyMessage(
                        message=msg,
                        format=FM_UNDEFINED,
                        data=b"",
                    )

            elif msg.token == TK_CMP_GETDEPENDS:
                if required_services is not None and required_services != "":
                    msg.token = TK_SYS_ACK
                    self.MB_ReplyMessage(
                        message=msg,
                        format=FM_STR_ARRAY_PIPE,
                        data=required_services.encode(),
                    )
                else:
                    msg.token = TK_SYS_NAK
                    self.MB_ReplyMessage(
                        message=msg,
                        format=FM_UNDEFINED,
                        data=b"",
                    )

            elif msg.token == TK_CMP_START:
                msg.token = TK_SYS_ACK
                self.MB_ReplyMessage(
                    message=msg,
                    format=FM_UNDEFINED,
                    data=b"",
                )
                return SE_SUCCESS

    def MB_Finalize(self):
        if self.my_service is None:
            return

        self.running = False
        if sys.platform != "win32":
            self.receive_sock.shutdown(socket.SHUT_RDWR)

        self.receive_sock.close()
        self.receive_thread.join()

    def _receive_message(
        self,
        sock: socket.socket,
        from_service: MBService,
        end_time: int,
    ) -> MBMessage:

        header_size = 32
        header = receive(sock, header_size, end_time)
        if len(header) != header_size:
            raise MBException("Error receiving header.", SE_MB_RCVHDR)

        magic = int.from_bytes(header[0:4], byteorder="little")
        if magic != TCP_MSG_MAGIC_EXT:
            raise MBException("Received magic is wrong", SE_MB_RCVHDR)

        data_size = int.from_bytes(header[4:8], byteorder="little")
        token = int.from_bytes(header[8:12], byteorder="little")
        format = int.from_bytes(header[12:16], byteorder="little")
        headers_size = int.from_bytes(header[16:20], byteorder="little")
        ttl = int.from_bytes(header[24:32], byteorder="little", signed=True)
        align_data = header[20:24]
        if int.from_bytes(align_data, byteorder="little") != 0x0:
            raise MBException("Invalid alignment data.", SE_MB_RCVHDR)

        dest_service_size = receive(sock, 4, end_time)
        if len(dest_service_size) != 4:
            raise MBException("Error receiving destination service size.", SE_MB_RCVHDR)

        dest_service_size = int.from_bytes(dest_service_size, byteorder="little")
        dest_service_name = None
        if dest_service_size > 0:
            dest_service_name = receive(sock, dest_service_size, end_time)
            if len(dest_service_name) != dest_service_size:
                raise MBException("Error receiving destination service size.", SE_MB_RCVHDR)

            dest_service_name = dest_service_name.decode()

        encrypted_data = b""
        if data_size > 0:
            encrypted_data = receive(sock, data_size, end_time)
            if len(encrypted_data) != data_size:
                raise MBException("Error receiving destination service size.", SE_MB_RCVBODY)

        received_data = encrypt(data=encrypted_data)

        headers = b""
        if headers_size > 0:
            headers = receive(sock, headers_size, end_time)
            if len(headers) != headers_size:
                raise MBException("Error receiving destination service size.", SE_MB_RCVBODY)

        trailer_bytes = receive(sock, 4, end_time)
        if len(trailer_bytes) != 4:
            raise MBException("Error receiving message trailer.", SE_MB_RCVTRAILER)
        trailer = int.from_bytes(trailer_bytes, byteorder="little")
        if trailer not in (TCP_MSG_TRAILER, TCP_MSG_TRAILER_OW):
            raise MBException("Invalid message trailer.", SE_MB_RCVTRAILER)

        crc_bytes = receive(sock, 4, end_time)
        if len(crc_bytes) != 4:
            raise MBException("Error receiving message trailer.", SE_MB_RCVTRAILER)

        received_crc = int.from_bytes(crc_bytes, byteorder="little")
        crc_data = TCP_MSG_MAGIC_EXT.to_bytes(4, byteorder="little")
        crc_data += data_size.to_bytes(4, byteorder="little")
        crc_data += token.to_bytes(4, byteorder="little")
        crc_data += format.to_bytes(4, byteorder="little")
        calculated_crc = calc_crc(crc_data, 0xffffffff)
        if data_size > 0:
            calculated_crc = calc_crc(received_data, calculated_crc)

        if received_crc != calculated_crc:
            raise MBException("Message CRC does not match.", SE_MB_RCVTRAILER)

        return MBMessage(
            token=token,
            format=format,
            ttl=ttl,
            destname=dest_service_name,
            data=received_data,
            headers=headers,
            trailer=trailer,
            remotehost=from_service.address,
        )


    def _send_message(
        self,
        destination: MBService,
        token: int,
        format: int = 0,
        data: bytes = None,
        headers: bytes = None,
        timeout: int = -1,
        ttl: int = -1,
        one_way: bool = False,
    ) -> Optional[MBMessage]:

        logger.debug("Sending token [0x%08X] to [%s]", token, destination)
        if data is None:
            data = b""

        if headers is None:
            headers = b""

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket.gethostbyaddr(destination.address)
        try:
            linger_data = struct.pack('ii', 0, 0)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, linger_data)
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            if timeout > 0:
                sock.settimeout(timeout / 1000000)
            else:
                sock.settimeout(5)

            start = time()
            if timeout > 0:
                end_time = start + timeout / 1000000
            else:
                end_time = start + 2 ** 16

            sock.connect((destination.address, destination.port))
            logger.debug("Successfully connected")

            now = time()
            remaining = end_time - now
            if remaining <= 0:
                raise MBTimeout("Timed-out on MB_Connect.", SE_TIMEOUT)

            self._send_data(
                sock=sock,
                dest_service_name=destination.name,
                token=token,
                format=format,
                data=data,
                headers=headers,
                ttl=ttl,
                one_way=one_way,
                end_time=end_time,
            )
            logger.debug("Data sent")

            reply = self._receive_message(
                sock=sock,
                from_service=destination,
                end_time=end_time,
            )
            logger.debug("Reply received with token [0x%08X]", reply.token)
            return reply
        except MBTimeout:
            raise
        except socket.timeout:
            raise MBTimeout("Timed-out on MB_Connect.", SE_TIMEOUT)
        except Exception as ex:
            raise MBException("Unknown error sending message.", SE_MB_ERR) from ex
        finally:
            sock.close()

    def _send_data(
        self,
        sock: socket.socket,
        token: int,
        format: int,
        data: bytes,
        headers: bytes,
        ttl: int,
        one_way: bool,
        dest_service_name: Optional[str],
        end_time: int,
    ):
        data_to_send = self._build_data_to_send(
            token=token,
            format=format,
            data=data,
            headers=headers,
            ttl=ttl,
            one_way=one_way,
            dest_service_name=dest_service_name,
        )

        now = time()
        remaining = end_time - now
        sock.settimeout(remaining)
        sock.sendall(data_to_send)

    def _build_data_to_send(
        self,
        token: int,
        format: int,
        data: bytes,
        headers: bytes,
        ttl: int,
        one_way: bool,
        dest_service_name: Optional[str],
    ):
        header_data = bytearray()
        header_data += TCP_MSG_MAGIC_EXT.to_bytes(4, byteorder="little")
        header_data += len(data).to_bytes(4, byteorder="little")
        header_data += token.to_bytes(4, byteorder="little")
        header_data += format.to_bytes(4, byteorder="little")

        crc = calc_crc(header_data, 0xffffffff)
        if len(data) > 0:
            crc = calc_crc(data, crc)

        send_data = header_data
        send_data += len(headers).to_bytes(4, byteorder="little")
        send_data += 0x0.to_bytes(4, byteorder="little")
        send_data += ttl.to_bytes(8, byteorder="little", signed=True)
        service_name_bytes = b""
        if dest_service_name is not None:
            service_name_bytes = dest_service_name.encode()
        send_data += len(service_name_bytes).to_bytes(4, byteorder="little")
        send_data += service_name_bytes
        send_data += encrypt(data=data)
        send_data += headers
        send_data += (TCP_MSG_TRAILER_OW if one_way else TCP_MSG_TRAILER).to_bytes(4, byteorder="little")
        send_data += crc.to_bytes(4, byteorder="little")

        return send_data

    def _receiver(self):
        while self.running:
            try:
                sock, address = self.receive_sock.accept()
                logger.debug("Accepted connection from %s:%d", address[0], address[1])
                sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                linger_data = struct.pack('ii', 0, 0)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, linger_data)
                from_service = MBService(address[0], address[1])
                msg = self._receive_message(
                    sock=sock,
                    from_service=from_service,
                    end_time=int(time() + 5),
                )
                logger.debug("Received token [0x%08X]", msg.token)
                if msg.trailer == TCP_MSG_TRAILER_OW:
                    self._send_data(
                        sock=sock,
                        dest_service_name=None,
                        token=msg.token,
                        format=FM_UNDEFINED,
                        data=b"",
                        headers=b"",
                        ttl=-1,
                        one_way=False,
                        end_time=int(time() + 5),
                    )
                    sock.close()
                    pass
                else:
                    setattr(msg, "_socket", sock)

                try:
                    with self.global_queue_lock:
                        self.msg_queue.put(
                            item=msg,
                            block=False,
                        )
                except queue.Full:
                    logger.exception("Queue is full")
                    try:
                        self._send_data(
                            sock=sock,
                            dest_service_name=from_service.name,
                            token=TK_SYS_QUEUE_FULL,
                            format=FM_UNDEFINED,
                            data=b"",
                            headers=b"",
                            ttl=-1,
                            one_way=False,
                            end_time=int(time() + 5),
                        )
                        sock.close()
                    except (Exception,):
                        logger.exception("Error sending TK_SYS_QUEUE_FULL")

            except (Exception,):
                if self.receive_sock.fileno() == -1 and self.running:
                    logger.error("Socket closed but we are still running")
                elif self.receive_sock.fileno() != -1:
                    logger.exception("Error on receiver")


def encrypt(data: bytes):
    encrypted_data = bytearray()
    for b in data:
        encrypted_data += (b ^ 0x9B).to_bytes(1, byteorder="little")

    return encrypted_data

def receive(sock: socket.socket, size:int, end_time: int):
    remaining_size = size
    total_received = b""

    while True:
        now = time()
        remaining_time = int(end_time - now)
        sock.settimeout(remaining_time)
        try:
            received = sock.recv(remaining_size)
            if received == b"":
                return total_received

            total_received += received
            remaining_size -= len(received)
            if remaining_size <= 0:
                return total_received
        except socket.timeout:
            raise MBTimeout("Timed-out on MB_Connect.", SE_TIMEOUT)



from threading import Thread

from assertpy import (
    assert_that,
)
from msgbus import (
    MBContext,
    MBException,
    MBService,
    SE_MB_ERR,
    SE_SUCCESS,
    TK_CMP_GETDEPENDS,
    TK_CMP_GETSERVICES,
    TK_CMP_START,
    TK_CMP_TERM_NOW,
    TK_SYS_ACK,
    TK_SYS_NAK,
)


def test_correctly_process_all_hv_wait_start_messages_when_component_has_services_and_dependencies():
    hv_simulator_service = MBService(
        address="127.0.0.1",
        port=15001,
    )
    hv_simulator = MBContext(
        my_service=hv_simulator_service,
        name="Sender1",
    )

    service = MBService(
        address="127.0.0.1",
        port=15002,
    )
    mb_context = MBContext(
        my_service=service,
        name="Sender2",
    )

    required_services = "service2|service3"
    exported_services = "service1|service2"

    def hv_thread():
        reply = hv_simulator.MB_SendFullMessage(
            destination=service,
            token=TK_CMP_GETDEPENDS,
        )
        assert_that(reply.token).is_equal_to(TK_SYS_ACK)
        assert_that(reply.data.decode()).is_equal_to(required_services)

        reply = hv_simulator.MB_SendFullMessage(
            destination=service,
            token=TK_CMP_GETSERVICES,
        )
        assert_that(reply.token).is_equal_to(TK_SYS_ACK)
        assert_that(reply.data.decode()).is_equal_to(exported_services)

        reply = hv_simulator.MB_SendFullMessage(
            destination=service,
            token=TK_CMP_START,
        )
        assert_that(reply.token).is_equal_to(TK_SYS_ACK)

        hv_simulator.MB_Finalize()

    hv_thread = Thread(target=hv_thread)
    hv_thread.start()

    ret = mb_context.MB_WaitStart(
        exported_services=exported_services,
        required_services=required_services,
    )

    assert_that(ret).is_equal_to(SE_SUCCESS)

    hv_thread.join()

    mb_context.MB_Finalize()


def test_correctly_process_all_hv_wait_start_messages_when_component_has_no_services_nor_dependencies():
    hv_simulator_service = MBService(
        address="127.0.0.1",
        port=15001,
    )
    hv_simulator = MBContext(
        my_service=hv_simulator_service,
        name="Sender1",
    )

    service = MBService(
        address="127.0.0.1",
        port=15002,
    )
    mb_context = MBContext(
        my_service=service,
        name="Sender2",
    )

    def hv_thread():
        reply = hv_simulator.MB_SendFullMessage(
            destination=service,
            token=TK_CMP_GETDEPENDS,
        )
        assert_that(reply.token).is_equal_to(TK_SYS_NAK)

        reply = hv_simulator.MB_SendFullMessage(
            destination=service,
            token=TK_CMP_GETSERVICES,
        )
        assert_that(reply.token).is_equal_to(TK_SYS_NAK)

        reply = hv_simulator.MB_SendFullMessage(
            destination=service,
            token=TK_CMP_START,
        )
        assert_that(reply.token).is_equal_to(TK_SYS_ACK)

        hv_simulator.MB_Finalize()

    hv_thread = Thread(target=hv_thread)
    hv_thread.start()

    ret = mb_context.MB_WaitStart()
    assert_that(ret).is_equal_to(SE_SUCCESS)

    mb_context.MB_Finalize()

    hv_thread.join()


def test_raises_exception_when_component_is_terminating():
    hv_simulator_service = MBService(
        address="127.0.0.1",
        port=15001,
    )
    hv_simulator = MBContext(
        my_service=hv_simulator_service,
        name="Sender1",
    )

    service = MBService(
        address="127.0.0.1",
        port=15002,
    )
    mb_context = MBContext(
        my_service=service,
        name="Sender2",
    )

    def hv_thread():
        reply = hv_simulator.MB_SendFullMessage(
            destination=service,
            token=TK_CMP_TERM_NOW,
        )
        assert_that(reply.token).is_equal_to(TK_SYS_ACK)


        hv_simulator.MB_Finalize()

    hv_thread = Thread(target=hv_thread)
    hv_thread.start()

    try:
        mb_context.MB_WaitStart()
    except MBException as ex:
        assert_that(ex.errorcode).is_equal_to(SE_MB_ERR)

    mb_context.MB_Finalize()

    hv_thread.join()



import os
from typing import Optional

from ._MBContext import MBContext
from ._MBMessage import MBMessage
from ._MBService import MBService


class MBEasyContext(MBContext):
    def __init__(
        self,
        name: str,
        hv_service: Optional[MBService] = None,
        comp_service: Optional[MBService] = None,
    ):
        self.hv_service = hv_service
        if self.hv_service is None:
            address = os.environ["HVIP"]
            hv_port = int(os.environ["HVPORT"])

            self.hv_service = MBService(
                address=address,
                port=hv_port,
            )

        self.comp_service = comp_service
        if self.comp_service is None:
            address = os.environ["HVIP"]
            comp_port = int(os.environ["HVCOMPPORT"])
            self.comp_service = MBService(
                address=address,
                port=comp_port,
            )

        super().__init__(
            self.comp_service,
            name,
        )

    def MB_EasySendFullMessage(
        self,
        dest_name,
        token,
        format=0,
        data=None,
        headers=None,
        timeout=-1,
        ttl=-1,
    ) -> MBMessage:

        destination = super().MB_LocateService(
            hv_service=self.hv_service,
            service_name=dest_name,
        )
        destination.name = dest_name

        return super().MB_SendFullMessage(
            destination=destination,
            token=token,
            format=format,
            data=data,
            headers=headers,
            timeout=timeout,
            ttl=ttl,
        )

    def MB_EasySendMessage(
        self,
        dest_name: str,
        token: int,
        format: int = 0,
        data: Optional[bytes] = None,
        timeout: int = -1
    ) -> MBMessage:

        return self.MB_EasySendFullMessage(
            dest_name=dest_name,
            token=token,
            format=format,
            data=data,
            timeout=timeout,
        )

    def MB_EasyGetMessage(self, timeout:Optional[int]=None) -> MBMessage:
        return super().MB_GetMessage(timeout=timeout)

    def MB_EasyReplyMessage(self, message: MBMessage) -> None:
        return super().MB_ReplyMessage(message=message)

    def MB_EasyWaitStart(self, exported_services=None, required_services=None) -> int:
        return super().MB_WaitStart(
            exported_services=exported_services,
            required_services=required_services,
        )

    def MB_EasyEvtSubscribe(self, subject: str) -> None:
        super().MB_EvtSubscribe(
            hv_service=self.hv_service,
            subject=subject,
        )

    def MB_EasyEvtUnsubscribe(self, subject: str) -> None:
        super().MB_EvtUnsubscribe(
            hv_service=self.hv_service,
            subject=subject,
        )

    def MB_EasyEvtSend(
        self,
        subject: str,
        type: str,
        xml: str,
        synchronous: bool = False,
        sourceid: int = 0,
        queue: str = None,
        timeout: int = -1,
    ):
        return self.MB_EasyEvtSendFull(
            subject=subject,
            type=type,
            headers=None,
            xml=xml,
            synchronous=synchronous,
            sourceid=sourceid,
            queue=queue,
            timeout=timeout,
            ttl=-1,
        )

    def MB_EasyEvtSendFull(
        self,
        subject: str,
        type: str,
        headers: Optional[bytes],
        xml: str,
        synchronous: bool = False,
        sourceid: int = 0,
        queue: str = None,
        timeout: int = -1,
        ttl: int = -1,
    ):
        return super().MB_EvtSendFull(
            hv_service=self.hv_service,
            headers=headers,
            subject=subject,
            type=type,
            xml=xml,
            synchronous=synchronous,
            sourceid=sourceid,
            queue=queue,
            timeout=timeout,
            ttl=ttl,
        )

    def MB_EasyFinalize(self) -> None:
        super().MB_Finalize()

    def MB_EasyRegisterService(
        self,
        service_type,
        service_name
    ):
        return super().MB_RegisterService(
            my_service=self.comp_service,
            hv_service=self.hv_service,
            service_type=service_type,
            service_name=service_name,
        )



from ._MBContext import MBContext
from ._MBEasyContext import MBEasyContext
from ._MBException import MBException
from ._MBMessage import MBMessage
from ._MBService import MBService
from ._MBTimeout import MBTimeout
from ._format import (
    FM_ADDR_PORT,
    FM_IMG_GIF,
    FM_IMG_JPG,
    FM_IMG_PNG,
    FM_INT32,
    FM_INT64,
    FM_PARAM,
    FM_PDF,
    FM_STRING,
    FM_STR_ARRAY_PIPE,
    FM_SWAGGER,
    FM_UNDEFINED,
    FM_XML,
)
from ._status import (
    SE_MB_ERR,
    SE_MB_RCVBODY,
    SE_MB_RCVHDR,
    SE_MB_RCVTRAILER,
    SE_MB_RECEIVE,
    SE_NOTFOUND,
    SE_SUCCESS,
    SE_TIMEOUT,
)
from ._token import (
    TK_CMP_GETDEPENDS,
    TK_CMP_GETSERVICES,
    TK_CMP_RECREATE,
    TK_CMP_HEALTH_CHECK,
    TK_CMP_REGSERVICE,
    TK_CMP_RELOAD,
    TK_CMP_START,
    TK_CMP_TERM_NOW,
    TK_EVT_EVENT,
    TK_HV_SERVICEOFFLINE,
    TK_HV_SERVICEONLINE,
    TK_SYS_ACK,
    TK_SYS_HEALTH_CHECK,
    TK_SYS_NAK,
    TK_SYS_QUEUE_FULL,
)



from threading import Thread

from assertpy import (
    assert_that,
)
from msgbus import (
    FM_UNDEFINED,
    MBContext,
    MBService,
    TK_SYS_ACK,
)


def test_receives_only_the_message_within_specified_filters():
    sender_service1 = MBService(
        address="127.0.0.1",
        port=15001,
    )
    sender1 = MBContext(
        my_service=sender_service1,
        name="Sender1",
    )
    sender_service2 = MBService(
        address="127.0.0.1",
        port=15002,
    )
    sender2 = MBContext(
        my_service=sender_service2,
        name="Sender2",
    )
    sender_service3 = MBService(
        address="127.0.0.1",
        port=15003,
    )
    sender3 = MBContext(
        my_service=sender_service3,
        name="Sender3",
    )

    receiver_service = MBService(
        address="127.0.0.1",
        port=14581,
    )
    receiver = MBContext(
        my_service=receiver_service,
        name="Receiver",
    )

    def receiver_thread():
        msg = receiver.MB_GetMessage(filtermin=10, filtermax=20)
        assert_that(msg.token).is_equal_to(15)

        msg.token = TK_SYS_ACK
        receiver.MB_ReplyMessage(
            message=msg,
            format=FM_UNDEFINED,
            data=b"",
        )

        msg = receiver.MB_GetMessage()
        assert_that(msg.token).is_in(9, 21)

        msg.token = TK_SYS_ACK
        receiver.MB_ReplyMessage(
            message=msg,
            format=FM_UNDEFINED,
            data=b"",
        )

        msg = receiver.MB_GetMessage()
        assert_that(msg.token).is_in(9, 21)

        msg.token = TK_SYS_ACK
        receiver.MB_ReplyMessage(
            message=msg,
            format=FM_UNDEFINED,
            data=b"",
        )

        receiver.MB_Finalize()


    def sender_thread(sender: MBContext, token: int):
        reply = sender.MB_SendFullMessage(
            destination=receiver_service,
            token=token,
        )
        assert_that(reply.token).is_equal_to(TK_SYS_ACK)
        sender.MB_Finalize()

    receiver_thread = Thread(target=receiver_thread)
    sender1_thread = Thread(target=sender_thread, args=(sender1, 9))
    sender2_thread = Thread(target=sender_thread, args=(sender2, 21))
    sender3_thread = Thread(target=sender_thread, args=(sender3, 15))

    receiver_thread.start()
    sender1_thread.start()
    sender2_thread.start()
    sender3_thread.start()

    receiver_thread.join()
    sender1_thread.join()
    sender2_thread.join()
    sender3_thread.join()



from threading import Thread

import cmsgbus
import msgbus
from assertpy import assert_that


def test_send_message_from_python_is_received_on_msgbus():
    server_ip = "127.0.0.1"
    sender_port = 14581
    receiver_port = 14580
    dest_name = "a service"

    old_mb_context = cmsgbus.MBContext(
        my_service=cmsgbus.MBService(
            address=server_ip,
            port=receiver_port,
        ),
        name="old service",
    )

    new_mb_context = msgbus.MBContext(
        my_service=msgbus.MBService(
            address=server_ip,
            port=sender_port,
        ),
        name="new service",
    )

    send_and_receive(
        sender=new_mb_context,
        destination=msgbus.MBService(
            address=server_ip,
            port=receiver_port,
            name=dest_name,
        ),
        dest_name=dest_name,
        receiver=old_mb_context,
        server_ip=server_ip,
        message_count=100,
    )

    old_mb_context.MB_Finalize()
    new_mb_context.MB_Finalize()

def test_message_from_msgbus_is_received_on_python():
    server_ip = "127.0.0.1"
    sender_port = 14581
    receiver_port = 14580
    dest_name = "a service"

    old_mb_context = cmsgbus.MBContext(
        my_service=cmsgbus.MBService(
            address=server_ip,
            port=sender_port,
        ),
        name="old service",
    )

    new_mb_context = msgbus.MBContext(
        my_service=msgbus.MBService(
            address=server_ip,
            port=receiver_port,
        ),
        name="new service",
    )

    send_and_receive(
        sender=old_mb_context,
        destination=cmsgbus.MBService(
            address=server_ip,
            port=receiver_port,
            pservice=dest_name,
        ),
        dest_name=dest_name,
        receiver=new_mb_context,
        server_ip=server_ip,
        message_count=100,
    )

    old_mb_context.MB_Finalize()
    new_mb_context.MB_Finalize()


def test_message_from_python_is_received_on_python():
    server_ip = "127.0.0.1"
    sender_port = 14581
    receiver_port = 14580
    dest_name = "a service"

    old_mb_context = msgbus.MBContext(
        my_service=msgbus.MBService(
            address=server_ip,
            port=sender_port,
        ),
        name="old service",
    )

    new_mb_context = msgbus.MBContext(
        my_service=msgbus.MBService(
            address=server_ip,
            port=receiver_port,
        ),
        name="new service",
    )

    send_and_receive(
        sender=old_mb_context,
        destination=msgbus.MBService(
            address=server_ip,
            port=receiver_port,
            name=dest_name,
        ),
        dest_name=dest_name,
        receiver=new_mb_context,
        server_ip=server_ip,
        message_count=100,
    )

    old_mb_context.MB_Finalize()
    new_mb_context.MB_Finalize()

def send_and_receive(
    sender: msgbus.MBContext,
    destination: msgbus.MBService,
    dest_name: str,
    receiver: msgbus.MBContext,
    server_ip: str,
    message_count: int,
):
    sent_token = msgbus.TK_CMP_GETSERVICES
    sent_format = msgbus.FM_STRING
    sent_data = b"some data"
    sent_headers = b"some headers"
    sent_ttl = 30

    reply_token = msgbus.TK_SYS_ACK
    reply_format = msgbus.FM_STRING
    reply_data = b"reply_data"

    def receive_thread():
        for _ in range(message_count):
            received = receiver.MB_GetMessage()

            assert_that(received.token).is_equal_to(sent_token)
            assert_that(received.format).is_equal_to(sent_format)
            assert_that(received.ttl).is_equal_to(sent_ttl)
            assert_that(received.data).is_equal_to(sent_data)
            assert_that(received.headers).is_equal_to(sent_headers)
            assert_that(received.destname).is_equal_to(dest_name)
            assert_that(received.remotehost).is_equal_to(server_ip)

            received.token = reply_token
            receiver.MB_ReplyMessage(
                message=received,
                format=reply_format,
                data=reply_data,
            )

    Thread(target=receive_thread).start()

    for _ in range(message_count):
        reply = sender.MB_SendFullMessage(
            destination=destination,
            token=sent_token,
            format=sent_format,
            data=sent_data,
            headers=sent_headers,
            timeout=5000000000,
            ttl=sent_ttl,
        )

        assert_that(reply.token).is_equal_to(reply_token)
        assert_that(reply.format).is_equal_to(reply_format)
        assert_that(reply.data).is_equal_to(reply_data)
        assert_that(reply.remotehost).is_equal_to(server_ip)



from typing import Optional


class MBService:
    def __init__(
        self,
        address: str,
        port: int,
        name: Optional[str] = None,
    ):

        self.address = address
        self.port = port
        self.name = name

    def __str__(self):
        return f"{self.address}:{self.port} - {self.name}"
